from io import BytesIO

import pandas as pd
import requests
from cannerflow.logging import *
from cannerflow.utils import *

__all__ = ["Query"]


class Query(object):
    @staticmethod
    def __get_pd_by_parquet_file(url, engine):
        r = requests.get(url)
        return pd.read_parquet(BytesIO(r.content), engine=engine)

    def __init__(
        self,
        workspace_id,
        request,
        sql,
        cache_refresh,
        cache_ttl,
        data_format,
        fetch_by='storage',
    ):
        self.workspace_id = workspace_id
        self.request = request
        self.sql = sql
        self.data_format = data_format
        self.offset = 0
        self.limit = 5000000
        self._columns = None
        self.create_sql_query_payload = gen_create_sql_query_payload(
            workspace_id=workspace_id,
            cache_refresh=cache_refresh,
            cache_ttl=cache_ttl,
            sql=sql)
        self.fetch_by = fetch_by
        self.data = []
        self.row_count = 0
        self.id = None
        self.status = None
        self.error = None
        self.statement_id = None
        self.create_sql_query()

    def __ensure_row_count(self):
        if self.row_count is None:
            self.get_first()

    def __ensure_columns(self):
        if self._columns is None:
            self.get_first()

    def __iter__(self):
        index = 0
        page = 1
        page_size = 50000
        self.get_first()  # initial columns and row count

        # fetch first page
        result = self.request.post(gen_sql_result_payload(self.id, page_size, 0)).get(
            'sqlResultPagination')
        data = result['result']

        while (page - 1) * page_size + index < self.row_count:
            if index == len(data):
                # fetch new page
                result = self.request.post(gen_sql_result_payload(self.id, page_size, page_size * page)).get('sqlResultPagination')
                data = result['result']
                index = 0
                page += 1
            elif index < len(data):
                rtn_data = data_factory(data_format=self.data_format, columns=self.columns, data=[data[index]])
                yield rtn_data
                index += 1
            else:
                break

    @property
    def columns(self):
        self.__ensure_columns()
        return self._columns

    def create_sql_query(self):
        result = self.request.post(self.create_sql_query_payload).get('createSqlQuery')
        if result['status'] == 'FAILED':
            print(result)
            raise RuntimeError('Execute sql failed')
        self.update_info(result)

    def get_sql_query(self):
        result = self.request.post(gen_sql_query_payload(self.id)).get('sqlQuery')
        self.update_info(result)

    def delete_statement(self):
        if self.statement_id is None:
            return
        self.request.post(gen_delete_statement_payload(self.statement_id))
        # update info
        self.get_sql_query()

    def get_sql_result(self, limit, offset):
        if self.fetch_by is 'storage':
            self.__get_sql_result_by_storage(limit, offset)
        else:
            self.__get_sql_result_by_restful(limit, offset)

    def __get_sql_result_by_storage(self, limit, offset, engine='fastparquet'):
        limit = limit or self.limit
        offset = offset or self.offset
        query_id = self.id
        workspace_id = self.workspace_id
        urls = self.request.get(f'api/v1/query/{query_id}/result/signedUrls?workspaceId={workspace_id}').get('signedUrls')
        # get columns and row_count by api
        self.__get_sql_result_by_restful(0, 0)
        # show warning if exist nested collection.
        self.__check_nested_collection_and_warning()
        if len(urls) is 0:
            df = pd.DataFrame({})
        else:
            dfs = []
            count = 0
            for url in urls:
                df = self.__get_pd_by_parquet_file(url, engine)
                count += df.shape[0]
                dfs.append(df)
                # if we get enough data, stop getting the next one
                if count >= limit + offset:
                    df = pd.concat(dfs)
                    break
            self._columns = list(map(lambda column: {'name': column}, list(df.columns)))
        self.data = df.iloc[offset:limit + offset, :].to_numpy().tolist()

    def __get_sql_result_by_restful(self, limit, offset):
        limit = limit or self.limit
        offset = offset or self.offset
        result = self.request.post(gen_sql_result_payload(self.id, limit, offset)).get('sqlResultPagination')
        self._columns = result['columns']
        self.data = result['result']
        self.row_count = result['rowCount']

    def wait_for_finish(self, timeout=120, period=1):
        def check_status_and_update_info():
            if self.status != 'FINISHED':
                self.get_sql_query()
                return False
            else:
                return True
        wait_until(check_status_and_update_info, timeout, period)

    def update_info(self, result):
        self.id = result['id']
        self.status = result['status']
        self.error = result['error']
        self.row_count = result['rowCount']
        self.statement_id = result['statementId']

    def get_data(self):
        try:
            return data_factory(data_format=self.data_format, columns=self.columns, data=self.data)
        except Exception:
            raise RuntimeError(
                'Cannot get data correctly, please run query.wait_for_finish(timeout=seconds,period=seconds) first')

    def get_all(self):
        self.__ensure_row_count()
        self.get_sql_result(self.row_count, 0)
        return self.get_data()

    def get_first(self, limit=1):
        self.get_sql_result(limit, 0)
        return self.get_data()

    def get_last(self, limit=1):
        self.__ensure_row_count()
        offset = self.row_count - limit
        self.get_sql_result(limit, offset)
        return self.get_data()

    def get(self, limit, offset):
        self.get_sql_result(limit, offset)
        return self.get_data()

    def __check_nested_collection_and_warning(self):
        result = self.request.get(f'api/v1/query/{self.id}?workspaceId={self.workspace_id}')
        columns = result['columns']
        logger = get_logger('Query', log_level=logging.WARNING)
        for column in columns:
            if self.__is_collection(column['typeSignature']['rawType']):
                for eleColumn in column['typeSignature']['arguments']:
                    if eleColumn['kind'] == 'TYPE':
                        if self.__is_collection(eleColumn['value']['rawType']):
                            logger.warning(f"Nested Collection isn't supported in fetchBy-storage-mode. Column %s: %s "
                                           f"would be None.", column['name'], column['type'])
                    elif eleColumn['kind'] == 'NAMED_TYPE':
                        if self.__is_collection(eleColumn['value']['typeSignature']['rawType']):
                            logger.warning(f"Nested Collection isn't supported in fetchBy-storage-mode. Column %s: %s "
                                           f"would be None.", column['name'], column['type'])

    def __is_collection(self, type):
        if ['array', 'map', 'row'].__contains__(type):
            return True
        return False

    def kill(self):
        self.delete_statement()

    def get_iterrows(self):
        return self




from cannerflow.request import Request
from cannerflow.saved_query import SavedQuery
from cannerflow.query import Query
from cannerflow.file import File

__all__ = ["Workspace"]

class Workspace(object):
    def __init__(
        self,
        endpoint,
        headers,
        workspace_id,
        replaceLocalhostString
    ):
        self.replaceLocalhostString = replaceLocalhostString
        self.endpoint = self._replace_localhost(endpoint)
        self.headers = headers
        self.workspace_id = workspace_id
        request = Request(
            headers=headers,
            endpoint=self.endpoint
        )
        self.request = request
        self.saved_query = SavedQuery(
            request=request,
            workspace_id=workspace_id
        )
        self.file = File(
            request=request,
            workspace_id=workspace_id,
            replaceLocalhostString=replaceLocalhostString
        )
    def gen_query(self, sql, cache_refresh, cache_ttl, data_format, fetch_by):
        return Query(
            request=self.request,
            workspace_id=self.workspace_id,
            sql=sql,
            cache_refresh=cache_refresh,
            cache_ttl=cache_ttl,
            data_format=data_format,
            fetch_by=fetch_by
        )

    def _replace_localhost(self, url):
        if (self.replaceLocalhostString != None):
            return url.replace('localhost', self.replaceLocalhostString).replace('127.0.0.1', self.replaceLocalhostString)
        else:
            return url;


import requests
__all__ = ["Request"]

class Request():
    def __init__(
        self,
        endpoint,
        headers
    ):
        self._endpoint = endpoint
        self._headers = headers
    def get_graphql_url(self):
        return "{endpoint}/{path}".format(
            endpoint=self._endpoint,
            path='graphql'
        )
    def get_url(self, path):
        return "{endpoint}/{path}".format(
            endpoint=self._endpoint,
            path=path
        )
    def post(self, payload):
        http_response = {}
        try:
            http_response = requests.post(
                self.get_graphql_url(),
                json=payload,
                headers=self._headers
            )
            http_response.raise_for_status()
        except Exception as err:
            print(f'Error occured: {err}, {http_response.content}')
        else:
            data = http_response.json()
            if "error" in data:
                print(f'Error occured: {data.error}')
            return data.get('data')
    def get(self, path):
        http_response = {}
        try:
            http_response = requests.get(
                self.get_url(path),
                headers=self._headers
            )
            http_response.raise_for_status()
        except Exception as err:
            content = http_response.get('content')
            print(f'Error occured: {err}, {content}')
        else:
            data = http_response.json()
            if "error" in data:
                error = data.get('error')
                print(f'Error occured: {error}')
            return data




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import json
import csv
from io import StringIO
import pandas as pd
import cannerflow.workspace
import cannerflow.constants
import cannerflow.exceptions
import cannerflow.utils
import nest_asyncio
nest_asyncio.apply()

__all__ = ["bootstrap", "Client"]

def bootstrap(*args, **kwargs):
    return Client(*args, **kwargs)

class Client(object):
    def __init__(self,
        endpoint,
        workspace_id,
        headers=None,
        token=None,
        replaceLocalhostString=None
    ):
        endpoint = endpoint.rstrip('/')
        if (token != None):
            # token based
            self.workspace = cannerflow.workspace.Workspace(
                endpoint=endpoint,
                headers={
                    'Authorization': f'Token {token}'
                },
                workspace_id=workspace_id,
                replaceLocalhostString=replaceLocalhostString
            )
        else:
            self.workspace = cannerflow.workspace.Workspace(
                endpoint=endpoint,
                headers=headers,
                workspace_id=workspace_id,
                replaceLocalhostString=replaceLocalhostString
            )

    # SQL
    def use_saved_query(self, title, cache_refresh=False, cache_ttl=86400, data_format="list"):
        query = self.workspace.saved_query.get(title)
        return self.gen_query(
            sql=query['sql'],
            cache_refresh=cache_refresh,
            cache_ttl=cache_ttl,
            data_format=data_format)

    def list_saved_query(self):
        return self.workspace.saved_query.list_title()

    def gen_query(self, sql, cache_refresh=False, cache_ttl=86400, data_format="list", fetch_by="storage"):
        query = self.workspace.gen_query(sql, cache_refresh=cache_refresh, cache_ttl=cache_ttl, data_format=data_format, fetch_by=fetch_by)
        return query

    # files
    def list_file(self):
        return self.workspace.file.list_absolute_path()

    def get_binary(self, absolute_path):
        return self.workspace.file.get_content(absolute_path)

    def put_binary(self, absolute_path, data):
        return self.workspace.file.put(absolute_path, data)

    # wrappers
    def _get_csv_wrapper(self, absolute_path, encoding='utf-8'):
        return self.workspace.file.get_csv_wrapper(absolute_path, encoding=encoding)

    def _get_json_wrapper(self, absolute_path, encoding='utf-8'):
        return self.workspace.file.get_json_wrapper(absolute_path, encoding=encoding)

    def _get_image_wrapper(self, absolute_path):
        return self.workspace.file.get_image_wrapper(absolute_path)

    # csv
    def get_csv(self, absolute_path, encoding='utf-8'):
        return self._get_csv_wrapper(absolute_path, encoding).to_list()

    def get_pandas_csv(self, absolute_path, encoding='utf-8', *args, **kwargs):
        return self._get_csv_wrapper(absolute_path, encoding).to_pandas(*args, **kwargs)

    def put_csv(self, absolute_path, data, encoding='utf-8', *args, **kwargs):
        if isinstance(data, pd.DataFrame):
            return self.put_pandas_csv(absolute_path, data, encoding, *args, **kwargs)
        return self.put_csv_str(absolute_path, data, encoding, *args, **kwargs)

    def put_csv_str(self, absolute_path, data, encoding='utf-8', *args, **kwargs):
        str_io = StringIO()
        csv_writer = csv.writer(str_io, *args, **kwargs)
        csv_writer.writerows(data)
        value = str_io.getvalue().encode(encoding)
        return self.put_binary(absolute_path=absolute_path, data=value)

    def put_pandas_csv(self, absolute_path, data, encoding='utf-8'):
        data = data.to_csv(index=False).encode(encoding)
        return self.put_binary(absolute_path=absolute_path, data=data)

    # json
    def get_json(self, absolute_path, encoding='utf-8'):
        return self._get_json_wrapper(absolute_path, encoding).to_json()

    def get_pandas_json(self, absolute_path, encoding='utf-8', *args, **kwargs):
        return self._get_json_wrapper(absolute_path, encoding).to_pandas(*args, **kwargs)

    def put_json(self, absolute_path, data, encoding='utf-8'):
        if isinstance(data, pd.DataFrame):
            return self.put_pandas_json(absolute_path, data, encoding)
        else:
            return self.put_json_str(absolute_path, data, encoding)

    def put_json_str(self, absolute_path, data, encoding='utf-8'):
        data = json.dumps(data).encode(encoding)
        return self.put_binary(absolute_path=absolute_path, data=data)

    def put_pandas_json(self, absolute_path, data, encoding='utf-8'):
        data = data.to_json()
        self.put_json_str(absolute_path=absolute_path, data=data, encoding=encoding)

    # image
    def get_pil_image(self, absolute_path):
        return self._get_image_wrapper(absolute_path).to_pil()

    def get_np_image(self, absolute_path):
        return self._get_image_wrapper(absolute_path).to_np()



import requests
import csv
import pandas
from io import BytesIO

__all__ = ["CsvWrapper"]

class CsvWrapper(object):
    def __init__(
        self,
        content,
        encoding='utf-8'
    ):
        self.content = content
        self.encoding = encoding
    def to_list(self, delimiter=','):
        self.content_text = self.content.decode(self.encoding)
        cr = csv.reader(self.content_text.splitlines(), delimiter=delimiter)
        return list(cr)

    def to_pandas(self, *args, **kwargs):
        return pandas.read_csv(
            filepath_or_buffer=BytesIO(self.content),
            encoding=self.encoding,
            engine='python',
            *args,
            **kwargs
        )





class ParquetHandler(object):
    def __init__(self, request, ):
        self.request = request

    def get_result(self, workspace_id, object_name):
        signed_url = self.request.get(f'/api/getSignedUrl?workspaceId={workspace_id}&objectName={object_name}')
        data = self.request.get()



import requests
import json
import pandas
from io import BytesIO

__all__ = ["JsonWrapper"]

class JsonWrapper(object):
    def __init__(
        self,
        content,
        encoding='utf-8'
    ):
        self.content = content
        self.encoding = encoding
    def to_json(self):
        self.content_text = self.content.decode(self.encoding)
        return json.loads(self.content_text)
    def to_pandas(self, *args, **kwargs):
        return pandas.read_json(
            path_or_buf=BytesIO(self.content),
            *args,
            **kwargs
        )


__all__ = ["SavedQuery"]

class SavedQuery(object):
    def __init__(
        self,
        workspace_id,
        request
    ):
        self.workspace_id = workspace_id
        self.request = request
    def getPayload(self):
        return {
            'operationName': 'sqls',
            'query': """
                query sqls($where: SqlWhereInput!) {
                    sqls(where: $where) {
                        id
                        sql
                        title
                        description
                        workspaceId
                    }
                }
            """,
            'variables': {
                'where': {
                    'workspaceId': self.workspace_id
                }
            }
        }
    def fetch(self):
        return self.request.post(self.getPayload()).get('sqls')
    def list_title(self):
        data = self.fetch()
        return list(map(lambda x: x['title'], data))
    def get(self, title):
        queries = self.fetch()
        query = next(q for q in queries if q.get('title') == title)
        return  query




# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""

This module defines exceptions for Cannerflow operations. It follows the structure
defined in pep-0249.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import functools
import random
import time

import cannerflow.logging

logger = cannerflow.logging.get_logger(__name__)

class HttpError(Exception):
    pass


class Http503Error(HttpError):
    pass


class CannerflowError(Exception):
    pass


class TimeoutError(Exception):
    pass


class CannerflowQueryError(Exception):
    def __init__(self, error, query_id=None):
        self._error = error
        self._query_id = query_id

    @property
    def error_code(self):
        return self._error.get("errorCode", None)

    @property
    def error_name(self):
        return self._error.get("errorName", None)

    @property
    def error_type(self):
        return self._error.get("errorType", None)

    @property
    def error_exception(self):
        return self.failure_info.get("type", None) if self.failure_info else None

    @property
    def failure_info(self):
        return self._error.get("failureInfo", None)

    @property
    def message(self):
        return self._error.get("message", "Cannerflow did no return an error message")

    @property
    def error_location(self):
        location = self._error["errorLocation"]
        return (location["lineNumber"], location["columnNumber"])

    @property
    def query_id(self):
        return self._query_id

    def __repr__(self):
        return '{}(type={}, name={}, message="{}", query_id={})'.format(
            self.__class__.__name__,
            self.error_type,
            self.error_name,
            self.message,
            self.query_id,
        )

    def __str__(self):
        return repr(self)


class CannerflowExternalError(CannerflowQueryError):
    pass


class CannerflowInternalError(CannerflowQueryError):
    pass


class CannerflowUserError(CannerflowQueryError):
    pass


def retry_with(handle_retry, exceptions, conditions, max_attempts):
    def wrapper(func):
        @functools.wraps(func)
        def decorated(*args, **kwargs):
            error = None
            result = None
            for attempt in range(1, max_attempts + 1):
                try:
                    result = func(*args, **kwargs)
                    if any(guard(result) for guard in conditions):
                        handle_retry.retry(func, args, kwargs, None, attempt)
                        continue
                    return result
                except Exception as err:
                    error = err
                    if any(isinstance(err, exc) for exc in exceptions):
                        handle_retry.retry(func, args, kwargs, err, attempt)
                        continue
                    break
            logger.info("failed after {} attempts".format(attempt))
            if error is not None:
                raise error
            return result

        return decorated

    return wrapper


class DelayExponential(object):
    def __init__(
        self, base=0.1, exponent=2, jitter=True, max_delay=2 * 3600  # 100ms  # 2 hours
    ):
        self._base = base
        self._exponent = exponent
        self._jitter = jitter
        self._max_delay = max_delay

    def __call__(self, attempt):
        delay = float(self._base) * (self._exponent ** attempt)
        if self._jitter:
            delay *= random.random()
        delay = min(float(self._max_delay), delay)
        return delay


class RetryWithExponentialBackoff(object):
    def __init__(
        self, base=0.1, exponent=2, jitter=True, max_delay=2 * 3600  # 100ms  # 2 hours
    ):
        self._get_delay = DelayExponential(base, exponent, jitter, max_delay)

    def retry(self, func, args, kwargs, err, attempt):
        delay = self._get_delay(attempt)
        time.sleep(delay)


# PEP 249
class Error(Exception):
    pass


class Warning(Exception):
    pass


class InterfaceError(Error):
    pass


class DatabaseError(Error):
    pass


class InternalError(DatabaseError):
    pass


class OperationalError(DatabaseError):
    pass


class ProgrammingError(DatabaseError):
    pass


class IntegrityError(DatabaseError):
    pass


class DataError(DatabaseError):
    pass


class NotSupportedError(DatabaseError):
    pass



import numpy as np
from io import BytesIO
from PIL import Image

__all__ = ["ImageWrapper"]

class ImageWrapper(object):
    def __init__(
        self,
        content
    ):
        self.content = content
    def to_pil(self):
        return Image.open(BytesIO(self.content));

    def to_np(self, *args, **kwargs):
        return np.asarray(self.to_pil())


import requests
import uuid
import base64
import aiohttp
import asyncio
import traceback
import mimetypes
from cannerflow.csv_wrapper import CsvWrapper
from cannerflow.json_wrapper import JsonWrapper
from cannerflow.image_wrapper import ImageWrapper
from io import BytesIO, StringIO


__all__ = ["File"]

BLOCK_SIZE = 10485760 # 10MB


class File(object):
    def __init__(
        self,
        workspace_id,
        request,
        replaceLocalhostString
    ):
        self.workspace_id = workspace_id
        self.request = request
        self.replaceLocalhostString = replaceLocalhostString

    @staticmethod
    def _create_block_id():
        id = str(uuid.uuid4())
        return base64.b64encode(id.encode()).decode()

    @staticmethod
    def _get_payload(workspace_id):
        return {
            'operationName': 'files',
            'query': """
                query files($where: FileWhereInput!, $recursive: Boolean) {
                    files(where: $where, recursive: $recursive) {
                        name
                        absolutePath
                        size
                        isFolder
                        lastModified
                        wiki
                    }
                }
            """,
            'variables': {
                'where': {
                    'workspaceId': workspace_id
                },
                'recursive': False
            }
        }

    def fetch(self):
        return self.request.post(File._get_payload(self.workspace_id)).get('files')

    def list_absolute_path(self):
        data = self.fetch()
        return list(map(lambda x: '/' + x['absolutePath'], data))

    def _replace_localhost(self, url):
        if self.replaceLocalhostString is not None:
            return url.replace('localhost', self.replaceLocalhostString).replace('127.0.0.1', self.replaceLocalhostString)
        else:
            return url

    def _get_download_signed_url(self, absolute_path):
        object_name = absolute_path
        if absolute_path[0] == '/':
            object_name = absolute_path[1:]
        data = self.request.get(f'api/getSignedUrl?workspaceId={self.workspace_id}&objectName={object_name}')
        return data

    def _get_upload_signed_url(self, absolute_path):
        object_name = absolute_path
        if absolute_path[0] == '/':
            object_name = absolute_path[1:]
        data = self.request.get(f'api/putSignedUrl?workspaceId={self.workspace_id}&objectName={object_name}')
        return data

    def _get(self, absolute_path):
        data = self._get_download_signed_url(absolute_path)
        # quick fix: in notebook we can't get minio data from localhost
        url = self._replace_localhost(data['signedUrl'])
        http_response = {}
        try:
            http_response = requests.get(url)
            http_response.raise_for_status()
        except Exception as err:
            content = http_response.get('content')
            print(f'Error occurred: {err}, {content}')
        else:
            return http_response

    def get_content(self, absolute_path):
        r = self._get(absolute_path)
        return r.content

    def get_text(self, absolute_path, encoding='utf-8'):
        r = self._get(absolute_path)
        r.encoding = encoding
        return r.text

    def get_csv_wrapper(self, absolute_path, encoding='utf-8'):
        content = self.get_content(absolute_path)
        csv_wrapper = CsvWrapper(content=content, encoding=encoding)
        return csv_wrapper

    def get_json_wrapper(self, absolute_path, encoding='utf-8'):
        content = self.get_content(absolute_path)
        json_wrapper = JsonWrapper(content=content, encoding=encoding)
        return json_wrapper

    def get_image_wrapper(self, absolute_path):
        content = self.get_content(absolute_path)
        image_wrapper = ImageWrapper(content=content)
        return image_wrapper

    @staticmethod
    async def _put_azure_block(url, file):
        block_id = File._create_block_id()
        async with aiohttp.ClientSession() as session:
            async with session.put(url=f'{url}&comp=block&blockId={block_id}', data=file) as response:
                response.raise_for_status()
        return block_id

    @staticmethod
    def _put_azure_block_list(url, block_ids, headers):
        id_map = map(lambda id: f'<Latest>{id}</Latest>', block_ids)
        content = '\n'.join(id_map)
        xml = """<?xml version="1.0" encoding="utf-8"?>
<BlockList>
  {}
</BlockList>""".format(content)
        headers['Content-Type'] = 'application/xml'
        http_response = requests.put(f'{url}&comp=blocklist',
                                     data=xml,
                                     headers=headers
                                     )
        return http_response

    @staticmethod
    async def _put_azure(url, data, headers):
        blocks = [data[i:i+BLOCK_SIZE] for i in range(0, len(data), BLOCK_SIZE)]
        block_ids = await asyncio.gather(*[File._put_azure_block(url=url, file=BytesIO(block)) for block in blocks])
        http_response = File._put_azure_block_list(url=url, block_ids=block_ids, headers=headers)
        return http_response

    @staticmethod
    def _put(url, data, headers):
        return requests.put(url,
                            data=data,
                            headers=headers
                            )

    def put(self, absolute_path, data):
        res = self._get_upload_signed_url(absolute_path)
        # quick fix: in notebook we can't get minio data from localhost
        url: str = self._replace_localhost(res['signedUrl'])
        info_id = res['fileInfoId']
        http_response = {}
        headers = {}
        try:
            if url.find('.blob.core.windows.net') is not -1:
                headers['x-ms-meta-file_info_id'] = info_id
                loop = asyncio.get_event_loop()
                http_response = loop.run_until_complete(self._put_azure(url=url, data=data, headers=headers))
            else:
                content_type = mimetypes.guess_type(absolute_path)[0]
                headers['x-amz-meta-file_info_id'] = info_id
                headers['Content-Type'] = content_type
                http_response = self._put(url=url, data=data, headers=headers)
            http_response.raise_for_status()
        except Exception as err:
            content = http_response.content
            print(f'Error occurred when put object: {err}, {content}, {traceback.format_exc()}')
            return False
        return True



import pandas as pd
from numpy import array
import time

__all__ = [
  "wait_until",
  "gen_create_sql_query_payload",
  "gen_sql_query_payload",
  "gen_sql_result_payload",
  "gen_delete_statement_payload",
  "data_factory"
]

def wait_until(somepredicate, timeout, period=1, *args, **kwargs):
  mustend = time.time() + timeout
  while time.time() < mustend:
    if somepredicate(*args, **kwargs): return True
    time.sleep(period)
  raise RuntimeError(f'Executeing query exceed {timeout} seconds, please call query.wait_for_finish(timeout=int) to set a longer waiting time') 

def gen_create_sql_query_payload(sql, cache_refresh, cache_ttl, workspace_id):
    return {
        'operationName': 'createSqlQuery',
        'query': """
            mutation createSqlQuery($data: SqlQueryCreateInput!, $where: SqlQueryWhereInput!) {
                createSqlQuery(data: $data, where: $where) {
                    id
                    status
                    error
                    location
                    statementId
                    rowCount
                }
            }
        """,
        'variables': {
            'data': {
                'sql': sql,
                'cacheRefresh': cache_refresh,
                'cacheTTL': cache_ttl,
                'source': 'NOTEBOOK'
            },
            'where': {
                'workspaceId': workspace_id
            }
        }
    }

def gen_sql_query_payload(sql_query_id):
    return {
        'operationName': 'sqlQuery',
        'query': """
            query sqlQuery($where: SqlQueryWhereUniqueInput!) {
                sqlQuery(where: $where) {
                    id
                    status
                    error
                    location
                    statementId
                    rowCount
                }
            }
        """,
        'variables': {
            'where': {
                'id': sql_query_id
            }
        }
    }

def gen_sql_result_payload(sql_query_id, limit, offset):
    return {
        'operationName': 'sqlResultPagination',
        'query': """
            query sqlResultPagination($where: SqlResultPaginationWhereInput!) {
                sqlResultPagination(where: $where) {
                    result
                    columns
                    rowCount
                }
            }
        """,
        'variables': {
            'where': {
                'id': sql_query_id,
                'limit': limit,
                'offset': offset,
                'source': 'NOTEBOOK'
            }
        }
    }

def gen_delete_statement_payload(statement_id):
    return {
        'operationName': 'deleteStatement',
        'query': """
            mutation deleteStatement($where: StatementWhereUniqueInput!) {
                deleteStatement(where: $where) {
                    id
                }
            }
        """,
        'variables': {
            'where': {
                'id': statement_id
            }
        }
    }

def data_factory(data_format='list', **options):
    def get_df(columns, data):
        data_list = get_list(columns, data)
        return pd.DataFrame(data_list[1:], columns=data_list[0])
    def get_np(columns, data):
        return array(data)
    def get_list(columns, data, with_column=True):
        if (with_column == True):
            column_names = list(map(lambda column: column['name'], columns))
            return [column_names] + data
        return data
    
    if (data_format == 'list'):
        return get_list(**options)
    if (data_format == 'df'):
        return get_df(**options)
    if (data_format == 'np'):
        return get_np(**options)
    raise RuntimeError(f'Unexpected data_format {data_format}')



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import ast
import re
from setuptools import setup
import textwrap


_version_re = re.compile(r"__version__\s+=\s+(.*)")


with open("cannerflow/__init__.py", "rb") as f:
    version = str(
        ast.literal_eval(_version_re.search(f.read().decode("utf-8")).group(1))
    )


kerberos_require = ["requests_kerberos"]

all_require = [kerberos_require, "pyarrow", "fastparquet", "json_lines"]

tests_require = all_require + ["httpretty", "pytest", "pytest-runner"]

py27_require = ["ipaddress", "typing"]

setup(
    name="cannerflow-python-client",
    author="Cannerflow Team",
    author_email="contact@canner.io",
    version=version,
    url="https://github.com/canner/cannerflow-python-client",
    packages=["cannerflow"],
    package_data={"": ["LICENSE", "README.md"]},
    description="Client for the Cannerflow",
    long_description=textwrap.dedent(
        """
    Client for cannerflow.
    """
    ),
    license="Apache 2.0",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: Microsoft :: Windows",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Database :: Front-Ends",
    ],
    install_requires=["click", "requests", "six", "pandas", "Pillow", "aiohttp", "nest_asyncio"],
    extras_require={
        "all": all_require,
        "kerberos": kerberos_require,
        "tests": tests_require,
        ':python_version=="2.7"': py27_require,
    },
)
